from collections import UserList


class TasksList(UserList):
    pass
    # def status(self, debug=False):
    #     if debug:
    #         for task in self:
    #             print(f"Task {task.edsl_name}")
    #             print(f"\t DEPENDS ON: {task.depends_on}")
    #             print(f"\t DONE: {task.done()}")
    #             print(f"\t CANCELLED: {task.cancelled()}")
    #             if not task.cancelled():
    #                 if task.done():
    #                     print(f"\t RESULT: {task.result()}")
    #                 else:
    #                     print(f"\t RESULT: None - Not done yet")

    #         print("---------------------")
